@@ -1,5 +1,5 @@
 
-# $Id: Debug.pm,v 1.20 2003/05/17 17:06:23 oradb Exp $
+# $Id: Debug.pm,v 1.43 2003/07/14 10:00:17 oradb Exp $
 
 =head1 NAME
 
@@ -18,7 +18,7 @@ use DBI;
 use Term::ReadKey;
 
 use vars qw($VERSION);
-$VERSION = do { my @r = (q$Revision: 1.20 $ =~ /\d+/g); sprintf "%d."."%02d" x $#r, @r };
+$VERSION = do { my @r = (q$Revision: 1.43 $ =~ /\d+/g); sprintf "%d."."%02d" x $#r, @r };
 
 my $DEBUG = $ENV{Oracle_Debug} || 0;
 
@@ -50,6 +50,8 @@ the perl debugger.
 
 Please see the I<packages/header.sql> file for credits for the original B<db> PL/SQL.
 
+Developed against B<Probe version 2.4>
+
 =item oradb
 
 The Perl chunk implements a perl-debugger-like interface to the Oracle
@@ -66,7 +68,7 @@ it from a command line usage.
 
 Ignore any methods which are prefixed with an underscore (_)
 
-We use a special table B<rfi_oracle_debug> for our own purposes.
+We use a special B<oradb_table> for our own purposes.
 
 Set B<Oracle_Debug>=1 for debugging information.
 
@@ -86,15 +88,23 @@ sub new {
 	my $proto = shift;
 	my $class = ref($proto) ? ref($proto) : $proto;
 	my $self  = bless({
-		'_config'  => do 'scripts/config', #$h_conf,
-		'_connected' => 0,
-		'_dbh'       => {},
-		'_debugpid'  => '',
-		'_name'      => '',
-		'_primed'    => 0,
-		'_sessionid' => '',
-		'_targetpid' => '',
-		'_type'      => '',
+		'_config'		=> do 'scripts/config', # $h_conf,
+		'_connect'	=> {
+			'debugpid'	=> '',
+			'primed'    => 0,
+			'sessionid'	=> '',
+			'targetid'	=> '',
+			'connected' => 0,
+			'synched'   => 0,
+			'syncs'   	=> 7,
+		},
+		'_dbh'			=> {},
+		'_unit'			=> {
+			'owner'			=> '',
+			'type'			=> '',
+			'name'			=> '',
+			'namespace'	=> '',
+		},
 	}, $class);
 	$self->_prime;
 	# $self->log($self.' '.Dumper($self)) if $DEBUG;
@@ -119,7 +129,7 @@ sub _prime {
 	} else {
 		# $self->{_dbh} = $self->dbh;
 		$self->{_dbh}->{$$} = $self->_connect($h_ref);
-		$self->{_primed}++ if $self->{_dbh}->{$$};
+		$self->{_connect}{primed}++ if $self->{_dbh}->{$$};
 		$self->dbh->func(20000, 'dbms_output_enable');
 		$self->self_check();
 	}
@@ -158,7 +168,7 @@ sub _connect {
 		$h_conf->{datasrc},	$h_conf->{user}, $h_conf->{pass}, $h_conf->{params} 
 	) || $self->fatal("Can't connect to database: $DBI::errstr");
 
-	$self->{_connected}++;
+	$self->{_connect}{connected}++;
 	$self->log("connected: $dbh") if $DEBUG;
 
 	return $dbh; #$id eq 'Debug' ? $dbh : 1;
@@ -175,8 +185,10 @@ Get a row
 sub getarow {
 	my $self  = shift;
 	my $sql   = shift;
+	my @res;
 
-	my @res = $self->dbh->selectrow_array($sql);
+	eval { @res = $self->dbh->selectrow_array($sql) };
+#	my @res = $self->dbh->selectrow_array($sql) || $self->error("failed <$sql>");
 	
 	if ($DEBUG) {
 		$self->log("failed to getarow: $sql $DBI::errstr") unless @res >= 1;
@@ -185,31 +197,104 @@ sub getarow {
 	return @res;
 }
 
+=item getahash
+
+Get a list of hashes
+
+	my ($res) = $o_debug->getahash($sql);
+
+=cut
+
+sub getahash {
+	my $self  = shift;
+	my $sql   = shift;
+	my @res;
+
+	eval { @res = $self->dbh->selectrow_hash($sql) };
+#	my @res = $self->dbh->selectrow_array($sql) || $self->error("failed <$sql>");
+	
+	if ($DEBUG) {
+		$self->log("failed to getahash: $sql $DBI::errstr") unless @res >= 1;
+	}
+
+	return @res;
+}
+
+
 # =============================================================================
 # parse and control
 # =============================================================================
 
 my %HISTORY = ();
-my $COMMANDS= 'rc|src|test|perl|sql|shell';
-my %GROUPS  = (
-	+0	=> [qw(rc src)],
-	+2	=> [qw(b c n r s)],
-	+4	=> [qw(l p)],
-	+6	=> [qw(h H ! q)],
-	+8	=> [qw(test perl sql shell)],
+my %TYPES   = (
+	'CU' => 'CURSOR',
+	'FU' => 'FUNCTION',
+	'PA' => 'PACKAGE',
+	'PR' => 'PROCEDURE',
+	'TR' => 'TRIGGER',
+	'TY' => 'TYPE',
 );
+my %NAMESPACES = (
+	'BO' => 'Namespace_pkg_body', 
+	'CU' => 'Namespace_cursor',
+	'FU' => 'Namespace_pkgspec_or_toplevel', 
+	'PA' => 'Namespace_pkgspec_or_toplevel', 
+	'PK' => 'Namespace_pkgspec_or_toplevel', 
+	'PR' => 'Namespace_pkgspec_or_toplevel', 
+	'SP' => 'Namespace_pkgspec_or_toplevel', 
+	'TR' => 'Namespace_trigger',
+);
+my %GROUPS  = (
+	+0	=> [qw()],
+	+1	=> [qw(b c n r s)],
+	+3	=> [qw(l L v T)],
+	+5	=> [qw(h H ! q)],
+	+6	=> [qw(context err perl rc sync sql shell info)],
+	+8	=> [qw(abort ping check test is_running)],
+);
+my $COMMANDS= join('|', @{$GROUPS{1}}, @{$GROUPS{3}}, @{$GROUPS{5}}, @{$GROUPS{6}}, @{$GROUPS{8}});
 my %COMMAND = (
+	'abort'		=> {
+		'handle'	=> 'abort',
+		'syntax'	=> 'abort',
+		'simple'	=> 'abort target', 
+		'detail'	=> 'abort currently running program in target session',
+	},
 	'b'		=> {
 		'handle'	=> 'break',
 		'syntax'	=> 'b [lineno]',
 		'simple'	=> 'set breakpoint', 
-		'detail'	=> 'set breakpoint on given line of code identified by name',
+		'detail'	=> 'set breakpoint on given line of code identified by unit name',
 	},
 	'c'	  => {
 		'handle'	=> 'continue',
 		'syntax'	=> 'c',
-		'simple'	=> 'continue to next interesting event',
-		'detail'	=> 'breakpoint or similar',
+		'simple'	=> 'continue',
+		'detail'	=> 'continue to breakpoint or other reason to stop',
+	},
+	'check'=> {
+		'handle'	=> 'self_check',
+		'syntax'	=> 'check',
+		'simple'	=> 'run a self_check',
+		'detail'	=> 'run a self_check against dbms_debug and probe communications',
+	},
+	'context'	  => {
+		'handle'	=> 'runtime', # context
+		'syntax'	=> 'context key[=val] [key[=val]]+',
+		'simple'	=> 'get/set context',
+		'detail'	=> 'get/set context for this instance: unit name, type, namespace etc.',
+	},
+	'err'	  => {
+		'handle'	=> 'plsql_errstr',
+		'syntax'	=> 'err',
+		'simple'	=> 'print plsql_errstr',
+		'detail'	=> 'display the DBI->plsql_errstr (if set)',
+	},
+	'info'	  => {
+		'handle'	=> 'info',
+		'syntax'	=> 'info',
+		'simple'	=> 'info on current environment',
+		'detail'	=> 'display information on current programs and db(NYI)',
 	},
 	'h'	  => {
 		'handle'	=> 'help',
@@ -224,8 +309,14 @@ my %COMMAND = (
 		'detail'	=> 'history listing not including single character commands',
 	},
 	'l'	  => {
+		'handle'	=> 'list_source',
+		'syntax'	=> 'l unitname [PROC|PACK|TRIG|...]',
+		'simple'	=> 'list source code',
+		'detail'	=> 'list source code given with library type',
+	},
+	'L'	  => {
 		'handle'	=> 'list_breakpoints',
-		'syntax'	=> 'l',
+		'syntax'	=> 'L',
 		'simple'	=> 'list breakpoints',
 		'detail'	=> 'on which line breakpoints exist',
 	},
@@ -235,12 +326,6 @@ my %COMMAND = (
 		'simple'	=> 'next line',
 		'detail'	=> 'continue until the next line',
 	},
-	'p'	  => {
-		'handle'	=> 'get_val',
-		'syntax'	=> 'p',
-		'simple'	=> 'print',
-		'detail'	=> 'print the value of a variable',
-	},
 	'perl'=> {
 		'handle'	=> 'perl',
 		'syntax'	=> 'perl <valid perl command>',
@@ -261,10 +346,9 @@ my %COMMAND = (
 	},
 	'rc'  => {
 		'handle'	=> 'recompile',
-		'syntax'	=> 'rc name [<PROC>(EDURE)|PACK(AGE)]',
-		'simple'	=> 'recompile + sync',
-		'detail'	=> 'recompile the program and synchronize with the target, '.
-                 '(note that this session _should_ hang until the procedure is executed in the target session)'
+		'syntax'	=> 'rc unitname',
+		'simple'	=> 'recompile',
+		'detail'	=> 'recompile the program/s given ',
 	},
 	's'	  => {
 		'handle'	=> 'step',
@@ -284,17 +368,42 @@ my %COMMAND = (
 		'simple'	=> 'SQL select',
 		'detail'	=> 'execute a SQL SELECT statement',
 	},
-	'src' => {
-		'handle'	=> 'list_source',
-		'syntax'	=> 'sql name [<PROC>(EDURE)|PACK(AGE)]',
-		'simple'	=> 'list source code',
-		'detail'	=> 'list source for given code (name)',
+	'sync'	  => {
+		'handle'	=> 'sync',
+		'syntax'	=> 'sync',
+		'simple'	=> 'sync',
+		'detail'	=> 'syncronize the sessions - '.
+                 '(note that this session _should_ hang until the procedure is executed in the target session)'
 	},
 	'test'=> {
-		'handle'	=> 'is_running',
+		'handle'	=> 'test',
 		'syntax'	=> 'test',
-		'simple'	=> 'target is running',
-		'detail'	=> 'test whether target session is currently running and responding',
+		'simple'	=> 'ping and check and if target is running',
+		'detail'	=> 'ping, run a self_check and test whether target session is currently running and responding',
+	},
+	'is_running'=> {
+		'handle'	=> 'is_running',
+		'syntax'	=> 'is_running',
+		'simple'	=> 'check target is_running',
+		'detail'	=> 'check whether target session is currently running and responding',
+	},
+	'ping'=> {
+		'handle'	=> 'ping',
+		'syntax'	=> 'ping',
+		'simple'	=> 'ping target',
+		'detail'	=> 'ping target session',
+	},
+	'T'=> {
+		'handle'	=> 'backtrace',
+		'syntax'	=> 'T',
+		'simple'	=> 'display backtrace',
+		'detail'	=> 'backtrace listings',
+	},
+	'v'	  => {
+		'handle'	=> 'value',
+		'syntax'	=> 'v varname[=value]',
+		'simple'	=> 'get/set variable',
+		'detail'	=> 'get or set the value of a variable, (use double quotes to contain spaces)',
 	},
 	'!'   => {
 		'handle'	=> 'rerun',
@@ -302,6 +411,12 @@ my %COMMAND = (
 		'simple'	=> 'run history command',
 		'detail'	=> 'run a command from the history list',
 	},
+	'x'   => {
+		'handle'	=> 'execute',
+		'syntax'	=> 'x sql',
+		'simple'	=> 'execute sql command',
+		'detail'	=> 'execute a sql command in the target session',
+	},
 );
 
 =cut
@@ -329,13 +444,26 @@ sub help {
 			$help .= "\tsyntax: $COMMAND{$levl}{syntax}\n\t$COMMAND{$levl}{detail}\n";
 	} else {
 		$levl = 'simple' unless $levl =~ /^(simple|detail|syntax|handle)$/io;
-		$help = "oradb help:\n";
+		my (@help, @left, @right) = ();
 		foreach my $grp (sort { $a <=> $b } keys %GROUPS) {
 			foreach my $char (@{$GROUPS{$grp}}) {
-				$help .= "\t".($levl ne 'syntax' ? "$char\t" : '')."$COMMAND{$char}{$levl}\n";
+				# $help .= "\t".($levl ne 'syntax' ? "$char\t" : '')."$COMMAND{$char}{$levl}\n";
+				my $myhelp = '    '.($levl ne 'syntax' ? sprintf('%-10s', $char) : '').($COMMAND{$char}{$levl}||'');
+				if ($grp =~ /^[13579]$/) {
+					push(@left, $myhelp);
+				} else {
+					push(@right, $myhelp);
+				}
 			}
-			$help .= "\n";
 		}
+		$#left = $#right if $#left < $#right;
+		$help = "oradb help:\n\n";
+		while (@left) {
+			no warnings; # empty right values
+			local $^W=0;
+			$help .= sprintf('%-45s', shift(@left) || '').shift(@right)."\n";
+		}
+		$help .= "\n";
 	}
 
 	return $help;
@@ -355,15 +483,15 @@ sub parse {
 	my $input= shift;
 
 	unless ($self->can($COMMAND{$cmd}{handle})) {
-		$self->error("command ($cmd) not understood");
+		$self->error("command '$cmd' not understood");
 		print $self->help;
 	} else {
+		$DB::single=2;
 		$DB::single=2;
 		my $handler = $COMMAND{$cmd}{handle} || 'help';
-		# print "xxx ->$handler<- xxx\n";
+		$self->log("cmd($cmd) input($input) handler($handler)") if $DEBUG;
 		my @res = $self->$handler($input);
-		# print "xxx ->@res<- xxx\n";
-		$self->log("cmd($cmd) input($input) handler($handler) returned(@res)") if $DEBUG;
+		$self->log("handler($handler) returned(@res)") if $DEBUG;
 		print @res;
 	}
 }
@@ -383,21 +511,24 @@ Wrapper for oradb->dbh->do() - internally we still use prepare and execute.
 sub do {
 	my $self = shift;
 	my $exec = shift;
-
-	my $csr  = $self->dbh->prepare($exec);
-	$self->fatal("Failed to prepare $exec - $DBI::errstr\n") unless $csr;
-
 	my $i_res;
-	eval {
-		$i_res = $csr->execute; # returning 0E0 is true/ok/good
-	};
 
-	if ($@) {
-		$self->fatal("Failure: $@ while evaling $exec - $DBI::errstr\n");
-	}
+	$self->log("*** incoming pl/sql: self($self) $exec args(@_)") if $DEBUG;
+	my $csr  = $self->dbh->prepare($exec);
+	unless ($csr) {
+		$self->error("Failed to prepare $exec - $DBI::errstr\n") unless $csr;
+	} else {
+		eval {
+			($i_res) = $csr->execute; # returning 0E0 is true/ok/good
+		};
 
-	unless ($i_res) {
-		$self->fatal("Failed to execute $exec - $DBI::errstr\n");
+		if ($@) {
+			$self->error("Failure: $@ while evaling $exec - $DBI::errstr\n");
+		}
+
+		unless ($i_res) {
+			$self->error("Failed to execute $exec - $DBI::errstr\n");
+		}
 	}
 
 	$self->log("do($exec)->res($i_res)") if $DEBUG;
@@ -407,9 +538,9 @@ sub do {
 
 =item recompile
 
-Recompile this procedure|function|package for debugging
+Recompile these procedure|function|package's for debugging
 
-	my $i_res = $oradb->recompile('PROCEDURE x');
+	$oradb->recompile('xsource');
 
 =cut
 
@@ -418,25 +549,89 @@ sub recompile {
 	my $args = shift;
 	my @res  = ();
 
-	my ($name, $type) = split(/\s+/, $args);
-	$name = uc($name); $type = '' unless $type;
-	$type = ($type =~ /^PROC/io ? 'PROCEDURE' : $type =~ /^FUNC/io ? 'FUNCTION' : $type =~ /^PACK/io ? 'PACKAGE' : 'PROCEDURE');
-	unless ($name =~ /^\w+$/o && $type =~ /^\w+$/o) {
-		$self->error("recompile requires a name($name) and type($type)");
-	} else {
-		$self->{_name} = $name;
-		$self->{_type} = $name;
-		my $exec = qq|ALTER $type $name COMPILE Debug|; 
-		$exec .= ' BODY' if $type eq 'PACKAGE';
-		@res = $self->do($exec)->get_msg;
-		print "Synching - once this hangs, execute this in the target session\n"; 
-		print "\t(if this does not hang, check the connection (with 'test'), and retry)\n";
-		@res = $self->sync;
+	my @names = split(/\s+/, $args);
+	foreach my $name (@names) {
+		my %data = $self->unitdata('name'=>$name);
+		if ($data{name} && $data{type}) {
+				$data{type} =~ s/BODY//;
+				my $exec = qq|ALTER $data{type} $data{name} COMPILE Debug|; $exec .= ' BODY' if $data{type} =~ /^PACKAGE|TYPE$/o;
+				my @msg = $self->do($exec)->get_msg;
+				print (@msg >= 1 ? "$data{name} recompiled\n" : "$data{name} failed recompilation!\n");
+				push(@res, @msg);
+		}
 	}
 
 	return @res;
 }
 
+=item synchronize
+
+Synchronize the debug and target sessions
+
+	$o_oradb->synchronize;
+
+=cut
+
+sub xsynchronize {
+	my $self = shift;
+	my $args = shift;
+	my @res  = ();
+
+	print "Synching - once this hangs, execute the code in the target session\n"; 
+	print "\t(if this does not hang, (it SHOULD), check the connection (with 'test'), and retry)\n";
+	@res = $self->sync;
+	$self->{_connect}{synched}++;
+	# print "Synched (if we hung - above - setting some breakpoints might be an idea...\n";
+
+	return @res;
+}
+
+=item unitdata
+
+Retrieve data for given unit - expects to recieve B<single> record from db!
+
+	%data = $o_oradb->unitdata('name'=>$name, 'type'=>$type, ...);
+
+=cut
+
+sub unitdata {
+	my $self = shift;
+	my %args = (
+		'name'	=> '',
+		'type'	=> '',
+		'owner'	=> '',
+	@_);
+	map { $args{$_} = '' unless $args{$_} } keys %args;
+	my %res  = ();
+
+	unless ($args{name} =~ /^\w+$/o) { # rjsf
+		$self->error("unit name($args{name}) is required");
+	} else {
+		my $sql = qq#SELECT DISTINCT(name || ':' || type || ':' || owner) FROM all_source 
+									WHERE UPPER(name) = UPPER('$args{name}')#;
+		$sql .= qq# AND UPPER(type) LIKE UPPER('$args{type}%')# if $args{type};
+		my ($data) = my @data = $self->getarow($sql);
+		my $input = join(', ', map { $_.'='.$args{$_} } sort keys %args);
+		unless (scalar(@data) == 1) {
+			$self->error("invalid or unambiguated data found via input($input)");
+		} else {
+			my ($name, $type, $owner) = split(':', $data);
+			unless ($name =~ /^\w+$/o) {
+				$self->error("invalid data($data) found via input($input)");
+			} else {
+				%res = (
+					'name'	=> $name, 
+					'type'	=> $type,
+					'owner'	=> $owner,
+				);
+				map { $self->{_unit}{lc($_)} = $res{$_} } keys %res;
+			} 
+		} 
+	}
+
+	return %res;
+}
+
 =item perl 
 
 Run a chunk of perl 
@@ -488,8 +683,8 @@ sub sql {
 		my $xsql = shift;
 		my @res  = ();
 
-		unless ($xsql =~ /^\s*SELECT\s+/o) {
-			$self->error("SELECT statements only please: $xsql");
+		unless ($xsql =~ /^\s*\w+\s+/io) {
+			$self->error("SQL statements only please: <$xsql>");
 		} else {
 			$xsql =~ s/\s*;\s*$//;
 			@res = ($self->getarow($xsql), "\n");
@@ -498,15 +693,15 @@ sub sql {
 		return @res;
 }
 
-=item run
+=item _run
 
 Run a chunk
 
-	$o_oradb->run($sql);
+	$o_oradb->_run($sql);
 
 =cut
 
-sub run {
+sub _run { # INTERNAL
       my $self = shift;
       my $xsql = shift;
 
@@ -536,17 +731,21 @@ sub target {
 	my $self = shift;
 
 	my $dbid = $self->start_target('rfi_oradb_sessionid');
-	
-	ReadMode 0;
-	print "orasql> enter a PL/SQL command to debug (debugger session must be running...)\n";
-	while (1) {
-		print "orasql>";
-		chomp(my $input = ReadLine(0));
-		$self->log("processing input($input)") if $DEBUG;
-		if ($input =~ /^\s*(q\s*|quit\s*)$/io) {
-			$self->quit;
-		} else {
-			$self->run($input);
+	if ($dbid) {
+		ReadMode 0;
+		print "orasql> enter a PL/SQL command to debug (debugger session must be running...)\n";
+		while (1) {
+			print "orasql>";
+			chomp(my $input = ReadLine(0));
+			$self->log("processing input($input)") if $DEBUG;
+			if ($input =~ /^\s*(q\s*|quit\s*)$/io) {
+				$self->quit;
+			} elsif ($input =~ /^\s*(h\s*|help\s*)$/io) {
+				print qq|No help menus for target session - simply enter code to debug (which will un-hang the debug session...)\n|;
+				$self->help;
+			} else {
+				$self->_run($input); 
+			}
 		}
 	}
 
@@ -565,11 +764,11 @@ sub start_target {
 	my $self = shift;
 	my $dbid = shift;
 
-	if ($self->{_debugid}) {
-		$self->fatal("mix-n-matching debug and target processes is not allowed");
+	if ($self->{_connect}{debugid}) {
+		$self->fatal("debug process may not run as a target instance");
 	}
 
-	$self->{_targetpid} = $dbid;
+	$self->{_connect}{targetpid} = $dbid;
 	my $x_res = $self->do('DELETE FROM '.$self->{_config}{table}); # currently we only allow a single session at a time
 
 	my $init = qq#
@@ -577,18 +776,21 @@ sub start_target {
 			xret VARCHAR2(32); 
 		BEGIN 
 			xret := dbms_debug.initialize('$dbid'); 
+			-- dbms_debug.debug_on(TRUE, FALSE); -- wait
+			dbms_debug.debug_on(TRUE, TRUE); -- immediate
 		END;
 	#;
 	$x_res = $self->do($init);
-
+=pod
 	my $ddid = qq#
 		BEGIN 
-			dbms_debug.debug_on(TRUE, FALSE); 
+			-- dbms_debug.debug_on(TRUE, FALSE); -- target releases debugger sync-hang by execute 
+			-- not certain the second TRUE is fully functional here...
+			dbms_debug.debug_on(TRUE, TRUE); -- debugger releases target hang with executes
 		END;
 		#; # should hang (if 2nd true) unless debugger running
 	$x_res = $self->do($ddid);
 
-=rjsf
 	# should be autonomous transaction
 	my $insert = qq#INSERT INTO $self->{_config}{table} 
            (created, debugpid, targetpid, sessionid, data) 
@@ -622,12 +824,15 @@ sub debugger {
 	my $i_cnt = 0;
 	while (1) {
 		print "oradb> ";
-		chomp(my $input = ReadLine(0));
+		chomp(my $input = ReadLine(0)); 
 		$self->log("processing command($input)") if $DEBUG;
-		if ($input =~ /^\s*($COMMANDS|.)\s*(.*)\s*$/o) {
-			my ($cmd, $args) = ($1, $2); $args =~ s/^\s+//o; $args =~ s/\s+$//o;
+		$input .= ' ';
+		if ($input =~ /^\s*($COMMANDS)\s+(.*)\s*$/o) {
+			my ($cmd, $args) = ($1, $2); 
+			$cmd =~ s/\s+$//; $args =~ s/^\s+//; $args =~ s/\s+$//;
 			$self->log("input($input) -> cmd($cmd) args($args)") if $DEBUG;
-			$HISTORY{++$i_cnt} = $cmd.' '.$args unless $cmd =~ /^\s*(.|!.*)\s*$/o;
+			my $res = $cmd.' '.$args;
+			$HISTORY{++$i_cnt} = $res unless $input =~ /^\s*(.|!.*)\s*$/o || grep(/^$res$/, map { $HISTORY{$_} } keys %HISTORY);
 			$self->parse($cmd, $args); # + process
 		} else {
 			$self->error("oradb> command ($input) not understood");	
@@ -651,10 +856,10 @@ sub start_debug {
 	my $pid  = shift;
 
 	# my $x_res = $self->do('UPDATE '.$self->{_config}{table}." SET debugpid = $pid");
-	if ($self->{_targetid}) {
-		$self->fatal("mix-n-matching target and debug processes is not allowed");
+	if ($self->{_connect}{targetid}) {
+		$self->fatal("target process may not run as a debug instance");
 	}
-	$self->{_debugpid} = $dbid;
+	$self->{_connect}{debugpid} = $dbid;
 
 	# SET serveroutput ON;                  -- done via dbi
 	my $x_res = $self->do(qq#ALTER session SET plsql_debug=TRUE#)->get_msg;
@@ -680,65 +885,50 @@ Blocks debug session until we exec in target session
 
 sub sync {
 	my $self = shift;
+	my @res  = ();
 
-=rjsf
+=pod rjsf
 	my ($tid) = $self->getarow('SELECT targetpid FROM '.$self->{_config}{table}." WHERE debugpid = '".$self->{_debugpid}."'");
 	$self->{_targetpid} = $tid;
 =cut
-
+	print "Synching - once this hangs, execute the code in the target session\n"; 
+	print "\t(if this does not hang, (it SHOULD), check the connection (with 'test'), and retry)\n";
+	
 	my $exec = qq#
 		DECLARE	
-			xret binary_integer;
+			xec     binary_integer;
 			runtime dbms_debug.runtime_info;
 		BEGIN	
-			xret := dbms_debug.synchronize(runtime);
-			IF xret = dbms_debug.success THEN
-				dbms_output.put_line('synched ' || runtime.program.name);
-			ELSIF xret = dbms_debug.error_timeout THEN
-				dbms_output.put_line('timed out');
-			ELSIF xret = dbms_debug.error_communication THEN
-				dbms_output.put_line('communication failure');
+			xec := dbms_debug.synchronize(runtime);
+			IF xec = dbms_debug.success THEN
+				NULL;
+				dbms_output.put_line('...synched ' || runtime.program.name);
 			ELSE
-				dbms_output.put_line('unknown error:' || xret);
+				dbms_output.put_line('Error: ' || oradb.errorcode(xec));
 			END IF;
 		END;
 	#;
 
-	return $self->do($exec)->get_msg;
+	my $test  = '';
+	my $i_cnt = 0;
+	while (1) {
+		$i_cnt++;
+		@res = $self->do($exec)->get_msg;
+		chomp($test = $self->is_running);
+		print ".";
+		last if ($i_cnt >= $self->{_connect}{syncs} || $test eq 'target is currently running');
+		sleep 1;
+	}
+	$self->{_connect}{synched}++;
+	print "\n$test\n";
+
+	return @res;
 }
 
 # ============================================================================= 
 # b c n s r exec
 # =============================================================================
 
-=item exec 
-
-Runs the given statement against the target session
-
-	my $i_res = $oradb->exec($call);
-
-=cut
-
-sub exec {
-	my $self = shift;
-	my $call = shift;
-	my $trim = $call; $trim =~ s/^(\w+)?\(.*$/$1/o;
-
-	$self->{_name} = $trim;
-	my @res = ();
-
-	# small loop
-	# check target is running
-	#
-	# alter call compile debug
-	# request target call this prog in 3 secs...
-	# sync (hang) with timeout
-	#
-	# $self->do($exec)->get_msg;
-
-	return @res; 
-}
-
 =item execute 
 
 Runs the given statement against the target session
@@ -747,15 +937,22 @@ Runs the given statement against the target session
 
 =cut
 
-sub _execute {
+sub execute {
 	my $self = shift;
 	my $xsql = shift;
 
+	$xsql =~ s/[\s\;]*$//;
+
 	my $exec = qq#
 		DECLARE 
-			col1 sys.dbms_debug_vc2coll; errm VARCHAR2(100); 
+			col1 sys.dbms_debug_vc2coll;
+			errm VARCHAR2(100);
 		BEGIN 
-			dbms_debug.execute($xsql, -1, 0, col1, errm); 
+			dbms_debug.execute('BEGIN $xsql; END;', 
+				-1, 0, col1, errm); 
+			IF (errm IS NOT NULL) THEN
+				DBMS_OUTPUT.put_line('Error($xsql): ' || errm);
+			END IF;
 		END;
 	#;
 
@@ -766,7 +963,7 @@ sub _execute {
 
 Set a breakpoint
 
-	my $i_res = $oradb->break("PROCNAME $i_line");
+	my $i_res = $oradb->break("$i_line $procedurename");
 
 =cut
 
@@ -775,17 +972,22 @@ sub break {
 	my $args = shift;
 	my @res  = ();
 
-	unless ($args =~ /^\s*(\w+)\s+(\d+)\s*$/o) {
-		$self->error("must supply the name($1) and line number($2) of a chunk of PL/SQL to set a breakpoint ($args)");
+	my ($line, $name) = split(/\s+/, $args);
+	# unless ($line =~ /^(\d+|\*)$/o) { <- fuzzy
+	unless ($line =~ /^(\d+)$/o) {
+		$self->error("must supply a valid line number($line) to set a breakpoint via($args)");
 	} else {
-		my ($name, $line, $owner) = (uc($1), $2, '');
-		my $exec = qq|
-			BEGIN 
-				db.b('$name', $line); 
-			END;
-		|;
-
-		@res = $self->do($exec)->get_msg;
+	  my $name = $name || $self->{_unit}{name} || '';
+		unless ($name =~ /^(\w+)$/o) { 
+			$self->error("library unit($name) must be given");
+		} else {
+			my $exec = qq|
+				BEGIN 
+					oradb.b('$name', $line); 
+				END;
+			|;
+			@res = $self->do($exec)->get_msg;
+		}
 	}
 
 	return @res;
@@ -804,7 +1006,7 @@ sub continue {
 
 	my $exec = qq#
 		BEGIN 
-    	db.continue_(dbms_debug.break_next_line);
+    	oradb.continue_(dbms_debug.break_any_call);
 		END;
 	#;
 
@@ -824,7 +1026,7 @@ sub next {
 
 	my $exec = qq#
 		BEGIN 
-    	db.continue_(dbms_debug.break_next_line);
+    	oradb.continue_(dbms_debug.break_next_line);
 		END;
 	#;
 
@@ -844,7 +1046,7 @@ sub step {
 
 	my $exec = qq#
 		BEGIN 
-    	db.continue_(dbms_debug.break_any_call);
+    	oradb.continue_(dbms_debug.break_any_call);
 		END;
 	#;
 
@@ -864,7 +1066,7 @@ sub return {
 
 	my $exec = qq#
 		BEGIN 
-    	db.continue_(dbms_debug.break_any_return);
+    	oradb.continue_(dbms_debug.break_return);
 		END;
 	#;
 
@@ -886,33 +1088,50 @@ Print runtime_info via dbms_output
 sub runtime {
 	my $self = shift;
 	my $sep = '-' x 80;
+	my @msg = ();
+
+	unless ($self->{_connect}{synched}) {
+		$self->error('not running yet');
+	} else {
+=pod
+   info_getStackDepth    CONSTANT PLS_INTEGER := 2;  -- get stack depth
+   info_getBreakpoint    CONSTANT PLS_INTEGER := 4;  -- get breakpoint number
+   info_getLineinfo      CONSTANT PLS_INTEGER := 8;  -- get program info
+   info_getOerInfo       CONSTANT PLS_INTEGER := 32; -- (Probe v2.4)
+=cut
 
 	my $exec = qq/
 		DECLARE 
 			runinfo dbms_debug.runtime_info; 
-			rsnt varchar2(40);
+			xinf BINARY_INTEGER DEFAULT dbms_debug.info_getBreakpoint + dbms_debug.info_getLineinfo + dbms_debug.info_getOerInfo;
+			xec  BINARY_INTEGER;
 		BEGIN 
-			-- rsnt := str_for_reason_in_runtime_info(runinfo.reason);
-			dbms_output.put_line('');
-			dbms_output.put_line('Runtime Info');
-    	dbms_output.put_line('  Name:       ' || runinfo.program.name);
-			dbms_output.put_line('  Line:          ' || runinfo.line#);
-			dbms_output.put_line('  Line:          ' || runinfo.program.line#);
-			dbms_output.put_line('  Terminated:    ' || runinfo.terminated);
-			-- dbms_output.put_line('  Reason         ' || rsnt);
+			xec := dbms_debug.get_runtime_info(xinf, runinfo);
+			IF xec = 0 THEN
+				dbms_output.put_line('Runtime Info:');
+				dbms_output.put_line('  Name:          ' || runinfo.program.name);
+				dbms_output.put_line('  Line:          ' || runinfo.line#);
+				dbms_output.put_line('  Owner:         ' || runinfo.program.owner);
+				dbms_output.put_line('  Unit:          ' || oradb.libunittype(runinfo.program.libunittype));
+				dbms_output.put_line('  Namespace:     ' || oradb.namespace(runinfo.program.namespace));
+			ELSE
+				dbms_output.put_line('   Error: ' || oradb.errorcode(xec));
+			END IF;
 		END;
 	/;
 
-	my @msg = $self->do($exec)->get_msg;
+		@msg = $self->do($exec)->get_msg;
+	}
 
 	return @msg >= 1 ? "\n".join("\n", $sep, @msg, $sep)."\n" : '...';
 }
 
+   
 =item backtrace 
 
 Print backtrace from runtime info via dbms_output
 
-	$oradb->backtrace();
+	$o_oradb->backtrace();
 
 =cut
 
@@ -921,22 +1140,23 @@ sub backtrace {
 
 	my $exec = qq#
 		DECLARE 
-			runinfo dbms_debug.runtime_info; 
+			tracing VARCHAR2(2000);
 		BEGIN 
-			db.print_backtrace(runinfo); 
+			dbms_debug.print_backtrace(tracing); 
+			dbms_output.put_line(tracing);
 		END;
 	#;
 
 	my @msg = $self->do($exec)->get_msg;
 
-	return @msg >= 1 ? @msg : '...';
+	return @msg;
 }
 
 =item list_source 
 
 Print source 
 
-	$oradb->list_source();
+	$oradb->list_source('xsource', [PROC|...]);
 
 =cut
 
@@ -945,17 +1165,16 @@ sub list_source {
 	my $args = shift;
 	my @res  = ();
 
-	my ($name, $type) = split(/\s+/, $args);
-	$name = uc($name); $type = '' unless $type;
-	$type = ($type =~ /^PROC/io ? 'PROCEDURE' : $type =~ /^FUNC/io ? 'FUNCTION' : $type =~ /^PACK/io ? 'PACKAGE BODY' : 'PROCEDURE');
-	unless ($name =~ /^\w+$/o && $type =~ /^\w+/o) {
-		$self->error("list source requires a name($name) and type($type)");
-	} else {
+	my ($name, $type) = split(/\s+/, $args); 
+	my %data = $self->unitdata('name'=>$name, 'type'=>$type);
+
+	if ($data{name} && $data{type}) {
 		my $exec = qq#
 			DECLARE
 				xsrc VARCHAR2(4000);
 				CURSOR src IS
-					SELECT line, text FROM all_source WHERE name = '$name' AND type = '$type' AND type != 'PACKAGE' ORDER BY name, line;
+					SELECT line, text FROM all_source WHERE name = '$data{name}' 
+					   AND type LIKE '$data{type}%' AND type != 'PACKAGE' ORDER BY name, line;
 			BEGIN
 				FOR rec IN src LOOP
 					xsrc := rec.line || ': ' || rec.text;
@@ -964,6 +1183,10 @@ sub list_source {
 			END;
 		#;
 		@res = $self->do($exec)->get_msg;
+		my $res = join('', @res);
+		unless ($res =~ /\w+/o) {
+			$self->error("no source($res) found with unit($data{name}) type($data{type})");
+		}
 	} 
 
 	return @res;
@@ -979,7 +1202,6 @@ Print breakpoint info
 
 sub list_breakpoints {
 	my $self = shift;
-	my $name = uc(shift) || $self->{_name};
 
 	my $exec = qq/
 		DECLARE
@@ -999,7 +1221,7 @@ sub list_breakpoints {
 	return $self->do($exec)->get_msg;
 }
 
-=rjsf
+=pod rjsf
 		vanilla version
 		DECLARE 
 			runinfo dbms_debug.runtime_info; 
@@ -1007,7 +1229,7 @@ sub list_breakpoints {
       i_after  number := 99;
       i_width  number := 80;
 		BEGIN 
-      db.print_runtime_info_with_source(runinfo, i_before, i_after, i_width);
+      oradb.print_runtime_info_with_source(runinfo, i_before, i_after, i_width);
 		END;
 =cut
 
@@ -1056,24 +1278,72 @@ sub rerun {
 # check and ping methods
 # =============================================================================
 
-=item self_check 
+=item info 
 
-Check the connections (fails otherwise)
+Info
 
-	my $i_ok = $oradb->self_check();
+	print $oradb->info;
 
 =cut
 
-sub self_check {
+sub info {
 	my $self = shift;
 
-	my $exec = qq#
-		BEGIN 
-			dbms_debug.self_check(10); 
-		END;
-	#;
+	my $src = $self->{_config}{datasrc} || '';
+	$src =~ s/^\w+:\w+://;
+	my @src = split(';', $src);
+	my %src = map { split('=', $_) } @src;
+	my ($probe, $version) = split(/:\s+/, $self->probe_version);
+	chomp($version);
 
-	return $self->do($exec) ? 1 : 0;
+	my %data = (
+		'host'			=> $src{host},
+		'instance'	=> uc($src{sid}),
+		'oradb'			=> $Oracle::Debug::VERSION,
+		'port'			=> $src{port},
+		'user'			=> $self->{_config}{user},
+		$probe			=> $version,
+	);
+	my ($i_max) = sort { $b <=> $a } map { length($_) } keys %data;
+
+	my @res = ("\n", (map { $_.(' 'x($i_max-length($_))).' = '.$data{$_}."\n" } sort keys %data), "\n");
+
+	return @res;
+}
+
+=item context
+
+Get and set context info
+
+	my $s_res = $o_oradb->context($name);         # get
+
+	my $s_res = $o_oradb->context($name, $value); # set
+
+=cut
+
+sub context {
+	my $self = shift;
+	my $args = shift || '';
+	my @args = my %args = ();
+	my @res  = ();
+
+	my ($i_max) = sort { $b <=> $a } map { length($_) } keys %{$self->{_unit}};
+
+	if (%args = ($args =~ /\G\s*(\w+)\s*=\s*(\w+)/go)) { # set
+		foreach (sort sort keys %args) {
+			my $call = "_$_";
+			push(@res, $_.(' 'x($i_max-length($_))).' = '.$self->$call($args{$_})."\n") if $self->can($call);
+		}
+	} elsif (@args = ($args =~ /\G\s*(\w+)\s*/go)) {     # get
+		foreach (sort @args) {
+			my $call = "_$_";
+			push(@res, $_.(' 'x($i_max-length($_))).' = '.$self->$call()."\n") if $self->can($call);
+		}
+	} else {                                             # all
+		@res = map { $_.(' 'x($i_max-length($_))).' = '.$self->{_unit}{$_}."\n" } sort keys %{$self->{_unit}};
+	}
+
+	return @res;
 }
 
 =item probe_version 
@@ -1100,6 +1370,44 @@ sub probe_version {
 	return $self->do($exec)->get_msg;
 }
 
+=item test 
+
+Call self_check, ping and is_running
+
+	my $i_ok = $oradb->test();
+
+=cut
+
+sub test {
+	my $self = shift;
+	my @res  = ();
+
+	push(@res, $self->self_check, $self->ping, $self->is_running);
+	
+	return @res;
+}
+
+=item self_check 
+
+Self->check
+
+	my $i_ok = $oradb->self_check; # 9.2
+
+=cut
+
+sub self_check {
+	my $self = shift;
+
+	my $exec = qq#
+		BEGIN 
+			dbms_debug.self_check(10);
+			dbms_output.put_line('checked');
+		END;
+		#;
+
+	return $self->do($exec)->get_msg;
+}
+
 =item ping 
 
 Ping the target process (gives an ORA-error if no target)
@@ -1113,7 +1421,7 @@ sub ping {
 
 	my $exec = qq#
 		BEGIN 
-			dbms_debug.ping;
+			dbms_debug.ping();
 			dbms_output.put_line('pinged');
 		END;
 		#;
@@ -1196,25 +1504,111 @@ sub get_msg {
 	return (@msg >= 1 ? join("\n", @msg)."\n" : "\n"); 
 }
 
-=item get_val
+=item value
 
-Get the value of a variable
+Get and set the value of a variable, in a procedure, or in a package
 
-	my $val = $o_debug->get_val($varname);
+	my $val = $o_oradb->value($name);
+
+	my $val = $o_oradb->value($name, $value);
 
 =cut
 
-sub get_val {
+sub value {
 	my $self = shift;
-	my $varn = shift;
+	my $args = shift || '';
+	my @res  = ();
+
+	my ($var, $getset) = ('', '', '');
+
+	if ($args =~ /^\s*(\w[\.\w]*)\s*:{0,1}=\s*(\S.+)?\s*$/o) {	# set
+		$var = "$1 := $2;";
+		$getset = '_set_val';
+	} elsif ($args =~ /^\s*(\w[\.\w]*)\s*$/) {        					# get
+		$var = $1;
+		$getset = '_get_val';
+	} else {																							# err
+		$self->error("unable to get or set variable - incorrect syntax: v $args");
+	}
+
+	if ($getset) {
+		@res = $self->$getset($var);
+	}
+
+	return @res;
+}
+
+=item _get_val
+
+Get the value of a variable
+
+	my $val = $o_debug->_get_val($varname);
+
+=cut
+
+sub _get_val {
+	my $self = shift;
+	my $xvar = shift;
 
 	my $exec = qq#
 		DECLARE
+			program dbms_debug.program_info;
+			runinfo dbms_debug.runtime_info; 
+			xinf BINARY_INTEGER DEFAULT dbms_debug.info_getBreakpoint + dbms_debug.info_getLineinfo + dbms_debug.info_getOerInfo;
+			xec  BINARY_INTEGER;
 			buff   VARCHAR2(500);
-			xret   BINARY_INTEGER;
 		BEGIN
-			xret := dbms_debug.get_value($varn, 0, buff, NULL);
-			dbms_output.put_line('value: ' || buff);
+			xec := dbms_debug.get_runtime_info(xinf, runinfo);
+			IF runinfo.program.namespace = 2 THEN 
+				/*
+					program := runinfo.program;
+	 				program.namespace  := dbms_debug.namespace_pkgspec_or_toplevel; -- as per docs...
+					program.Owner      := runinfo.program.owner;
+					program.Name       := runinfo.program.name;
+					xec := dbms_debug.get_value('$xvar', program, buff, NULL);
+				*/
+				xec := dbms_debug.get_value('$xvar', 0, buff, NULL);
+			ELSE
+				xec := dbms_debug.get_value('$xvar', 0, buff, NULL);
+			END IF; 
+			IF xec = dbms_debug.success THEN
+				dbms_output.put_line('$xvar = ' || buff);
+			ELSE
+				dbms_output.put_line('Error: ' || oradb.errorcode(xec));
+			END IF;
+		END;
+	#;
+
+	my @res = $self->do($exec)->get_msg;
+
+	return @res;
+}
+
+=item _set_val
+
+Set the value of a variable
+
+	my $val = $o_debug->_set_val($xset);
+
+=cut
+
+sub _set_val {
+	my $self = shift;
+	my $xset = shift;
+
+	# $self->error("unimplemented");
+
+	my $exec = qq#
+		DECLARE
+			xec BINARY_INTEGER;
+		BEGIN
+			xec := dbms_debug.set_value(0, '$xset');
+
+			IF xec = dbms_debug.success THEN
+				dbms_output.put_line('$xset succeeded');
+			ELSE
+				dbms_output.put_line('Error: ' || oradb.errorcode(xec));
+			END IF;
 		END;
 	#;
 	
@@ -1245,13 +1639,148 @@ sub audit {
 	return $res." $$";
 }
 
+# =============================================================================
+# get and put context methods
+# =============================================================================
+
+=item _check
+
+Return whether or not the given PLSQL target has a value of some sort
+
+	my $i_ok = $o_oradb->_check('unit');
+
+=cut
+
+sub _check {
+	my $self = shift;
+	my $targ = lc(shift);
+	my $i_ok = 0;
+	
+	unless ($targ =~ /^\w+$/o) {
+		$self->error("require a valid plsql target($targ) to check: ".join(', ', sort keys %{$self->{_unit}}));
+	} else {
+		$i_ok++ if $self->{_unit}{$targ} =~ /./o;
+	}
+
+	return $i_ok;
+}
+
+=item _unit
+
+Get and set B<unit> name for all consequent actions
+
+	$o_oradb->_unit;        # get
+
+	$o_oradb->_unit($name); # set
+
+=cut
+
+sub _unit {
+	my $self = shift;
+	my $args = shift || $self->{_unit}{name} || '';
+
+	unless ($args =~ /^\s*(\w+)\s*$/o) {
+		$self->error("valid alphanumeric unit($args) is required");
+	} else {
+		$self->{_unit}{name} = uc($args);
+	}
+	
+	$self->{_unit}{name};
+}
+
+=item _type 
+
+Get and set B<type> for all consequent actions
+
+	$o_oradb->_type;        # get
+
+	$o_oradb->_type($type); # set
+
+=cut
+
+sub _type {
+	my $self = shift;
+	my $args = shift || $self->{_unit}{type} || '';
+
+	my $xx = uc(substr($args, 0, 2));
+	unless ($TYPES{$xx} =~ /^(\w+)$/o) {
+		$self->error("invalid type($args) - the following are allowed: ".join(', ', sort VALUES %TYPES));
+	} else {
+		$self->{_unit}{type} = uc($1);
+	}
+	
+	$self->{_unit}{type};
+}
+
+=item _namespace
+
+Get and set B<unit> namespace for all consequent actions
+
+	$o_oradb->_namespace;         # get
+
+	$o_oradb->_namespace($space); # set
+
+=cut
+
+sub _namespace {
+	my $self = shift;
+	my $args = shift || $self->{_unit}{namespace} || '';
+
+	my $xx = uc(substr($args, 0, 2));
+	unless ($NAMESPACES{$xx} =~ /^(\w+)$/o) {
+		$self->error("invalid namespace($args) - the following are allowed: ".join(', ', sort VALUES %NAMESPACES));
+	} else {
+		$self->{_unit}{namespace} = uc($1);
+	}
+	
+	return $self->{_unit}{namespace};
+}
+
+=item _owner
+
+Get and set B<unit> owner for all consequent actions
+
+	$o_oradb->_owner;        # get
+
+	$o_oradb->_owner($user); # set
+
+=cut
+
+sub _owner {
+	my $self = shift;
+	my $args = shift || $self->{_unit}{owner} || '';
+
+	unless ($args =~ /^\s*(\w+)\s*$/o) {
+		$self->error("valid alphanumeric owner($args) is required");
+	} else {
+		$self->{_unit}{owner} = uc($1);
+	}
+	
+	return $self->{_unit}{owner};
+}
+
 # =============================================================================
 # error, log and cleanup methods
 # =============================================================================
 
+=item feedback 
+
+Feedback handler (currently just prints to STDOUT)
+
+	$o_debug->feedback("this");
+
+=cut
+
+sub feedback {
+	my $self = shift;
+	my $msgs = join(' ', @_);
+	print STDOUT 'ORADB> '."$msgs\n";
+	return $msgs;
+}
+
 =item log 
 
-Log handler (currently just prints to STDOUT)
+Log handler (currently just prints to STDERR)
 
 	$o_debug->log("this");
 
@@ -1260,7 +1789,7 @@ Log handler (currently just prints to STDOUT)
 sub log {
 	my $self = shift;
 	my $msgs = join(' ', @_);
-	print STDOUT '          '."$msgs\n";
+	print STDERR 'oradb: '."$msgs\n";
 	return $msgs;
 }
 
@@ -1274,7 +1803,8 @@ Quit the debugger
 
 sub quit {
 	my $self = shift;
-	print "oradb detaching\n";
+	$self->abort();
+	print "oradb detaching...\n";
 	# $self->detach;
 	exit;
 }
@@ -1287,9 +1817,10 @@ Error handler
 
 sub error {
 	my $self = shift;
-	# $DB::errstr;
+	$DB::errstr = $DB::errstr;
 	my $errs = join(' ', 'Error:', @_).($DB::errstr || '')."\n";
-	carp($errs);
+	print $errs;
+	# carp($errs);
 	return $errs;
 }
 
@@ -1304,6 +1835,32 @@ sub fatal {
 	croak(ref($self).' FATAL ERROR: ', @_);
 }
 
+=item abort 
+
+Tell the target session to abort the currently running program
+
+	$o_debug->abort;
+
+=cut
+
+sub abort {
+	my $self = shift;
+
+	my $exec = qq#
+		DECLARE 
+			runinfo dbms_debug.runtime_info; 
+			ret BINARY_INTEGER;
+		BEGIN 
+    	-- oradb.continue_(dbms_debug.abort_execution);
+    	ret := dbms_debug.continue(runinfo, dbms_debug.abort_execution, 0);
+		END;
+	#;
+	print "exec($exec)\n";
+
+	$self->do($exec)->get_msg;
+}
+
+
 =item detach
 
 Tell the target session to detach itself
@@ -1320,11 +1877,11 @@ sub detach {
 			dbms_debug.detach_session; 
 		END;
 	#;
-	$self->do->($exec)->get_msg;
+	$self->do($exec)->get_msg;
 
 	# autonomous transaction
-	# $self->do->('DELETE FROM '.$self->{_config}{table});
-	# $self->do->('COMMIT');
+	# $self->do('DELETE FROM '.$self->{_config}{table});
+	# $self->do('COMMIT');
 }
 
 sub DESTROY {
